#!/usr/bin/env python3
"""
Service Discovery Tool
Network service discovery and mapping tool.
Parameters:
    - target: Target network range (CIDR notation)
    - ports: Ports to scan for services
    - timeout: Timeout for service detection
Requires: nmap, zeroconf packages
"""
import nmap
import socket
import argparse
import logging
from datetime import datetime
import json
import threading
from zeroconf import ServiceBrowser, Zeroconf
import time
import queue
from concurrent.futures import ThreadPoolExecutor

class ServiceDiscovery:
    def __init__(self, target, ports=None, timeout=5):
        self.target = target
        self.ports = ports or "21-23,25,53,80,110,139,443,445,3306,3389,5432"
        self.timeout = timeout
        self.discovered_services = queue.Queue()
        self.nm = nmap.PortScanner()
        self.setup_logging()

    def setup_logging(self):
        logging.basicConfig(
            level=logging.INFO,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )
        self.logger = logging.getLogger(__name__)

    def scan_network(self):
        try:
            self.logger.info(f"Scanning network {self.target} for services...")
            self.nm.scan(self.target, self.ports, arguments='-sV -sS')
            
            for host in self.nm.all_hosts():
                for proto in self.nm[host].all_protocols():
                    ports = self.nm[host][proto].keys()
                    for port in ports:
                        service = self.nm[host][proto][port]
                        if service['state'] == 'open':
                            self.discovered_services.put({
                                'host': host,
                                'port': port,
                                'protocol': proto,
                                'service': service['name'],
                                'product': service.get('product', ''),
                                'version': service.get('version', ''),
                                'method': 'nmap'
                            })
        except Exception as e:
            self.logger.error(f"Error during network scan: {str(e)}")

    def discover_mdns_services(self):
        class MDNSListener:
            def __init__(self, parent):
                self.parent = parent

            def add_service(self, zeroconf, type, name):
                info = zeroconf.get_service_info(type, name)
                if info:
                    self.parent.discovered_services.put({
                        'host': socket.inet_ntoa(info.addresses[0]),
                        'port': info.port,
                        'protocol': 'mdns',
                        'service': type,
                        'name': name,
                        'properties': dict(info.properties),
                        'method': 'mdns'
                    })

        try:
            self.logger.info("Starting mDNS service discovery...")
            zeroconf = Zeroconf()
            listener = MDNSListener(self)
            browser = ServiceBrowser(zeroconf, "_http._tcp.local.", listener)
            browser = ServiceBrowser(zeroconf, "_https._tcp.local.", listener)
            browser = ServiceBrowser(zeroconf, "_printer._tcp.local.", listener)
            browser = ServiceBrowser(zeroconf, "_ipp._tcp.local.", listener)
            time.sleep(self.timeout)
            zeroconf.close()
        except Exception as e:
            self.logger.error(f"Error during mDNS discovery: {str(e)}")

    def discover_netbios_services(self):
        def query_netbios(host):
            try:
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(1)
                
                # NetBIOS Node Status Query
                query = (
                    b'\x82\x28\x00\x00\x00\x01\x00\x00\x00\x00\x00\x00'
                    b' CKAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\x00\x00\x21\x00\x01'
                )
                
                sock.sendto(query, (host, 137))
                data, _ = sock.recvfrom(1024)
                
                if data:
                    self.discovered_services.put({
                        'host': host,
                        'port': 137,
                        'protocol': 'netbios',
                        'service': 'NetBIOS',
                        'method': 'netbios'
                    })
            except:
                pass
            finally:
                sock.close()

        try:
            self.logger.info("Starting NetBIOS service discovery...")
            network = ipaddress.ip_network(self.target)
            with ThreadPoolExecutor(max_workers=50) as executor:
                executor.map(query_netbios, [str(ip) for ip in network.hosts()])
        except Exception as e:
            self.logger.error(f"Error during NetBIOS discovery: {str(e)}")

    def discover_snmp_services(self):
        def query_snmp(host):
            try:
                # SNMP GET request (community string: public)
                sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
                sock.settimeout(1)
                
                snmp_request = (
                    b'\x30\x26\x02\x01\x01\x04\x06\x70\x75\x62\x6c\x69\x63'
                    b'\xa0\x19\x02\x04\x6b\x8b\x44\x5b\x02\x01\x00\x02\x01'
                    b'\x00\x30\x0b\x30\x09\x06\x05\x2b\x06\x01\x02\x01\x05\x00'
                )
                
                sock.sendto(snmp_request, (host, 161))
                data, _ = sock.recvfrom(1024)
                
                if data:
                    self.discovered_services.put({
                        'host': host,
                        'port': 161,
                        'protocol': 'snmp',
                        'service': 'SNMP',
                        'method': 'snmp'
                    })
            except:
                pass
            finally:
                sock.close()

        try:
            self.logger.info("Starting SNMP service discovery...")
            network = ipaddress.ip_network(self.target)
            with ThreadPoolExecutor(max_workers=50) as executor:
                executor.map(query_snmp, [str(ip) for ip in network.hosts()])
        except Exception as e:
            self.logger.error(f"Error during SNMP discovery: {str(e)}")

    def save_results(self, output_file):
        services = []
        while not self.discovered_services.empty():
            services.append(self.discovered_services.get())

        results = {
            'timestamp': datetime.now().isoformat(),
            'target': self.target,
            'ports_scanned': self.ports,
            'services_found': len(services),
            'services': services
        }

        try:
            with open(output_file, 'w') as f:
                json.dump(results, f, indent=4)
            self.logger.info(f"Results saved to {output_file}")
        except Exception as e:
            self.logger.error(f"Error saving results: {str(e)}")

    def start_discovery(self, output_file=None):
        threads = []
        
        # Start network scan
        network_thread = threading.Thread(target=self.scan_network)
        network_thread.start()
        threads.append(network_thread)
        
        # Start mDNS discovery
        mdns_thread = threading.Thread(target=self.discover_mdns_services)
        mdns_thread.start()
        threads.append(mdns_thread)
        
        # Start NetBIOS discovery
        netbios_thread = threading.Thread(target=self.discover_netbios_services)
        netbios_thread.start()
        threads.append(netbios_thread)
        
        # Start SNMP discovery
        snmp_thread = threading.Thread(target=self.discover_snmp_services)
        snmp_thread.start()
        threads.append(snmp_thread)
        
        # Wait for all discoveries to complete
        for thread in threads:
            thread.join()
        
        if output_file:
            self.save_results(output_file)
        
        self.logger.info("\nDiscovery Summary:")
        services = []
        while not self.discovered_services.empty():
            services.append(self.discovered_services.get())
        
        self.logger.info(f"Total services discovered: {len(services)}")
        for method in ['nmap', 'mdns', 'netbios', 'snmp']:
            count = len([s for s in services if s['method'] == method])
            self.logger.info(f"{method.upper()} discovered services: {count}")

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Service Discovery Tool")
    parser.add_argument("target", help="Target network range (CIDR notation)")
    parser.add_argument("--ports",
                      help="Ports to scan (default: common ports)")
    parser.add_argument("--timeout", type=int, default=5,
                      help="Timeout for service detection")
    parser.add_argument("--output", help="Output file for results")
    args = parser.parse_args()

    discovery = ServiceDiscovery(args.target, args.ports, args.timeout)
    discovery.start_discovery(args.output)
